#include "lusol.h"

extern "C" void aplb_(int*, int*, int*, REAL*, int*, int*,
           REAL*, int*, int*, REAL*, int*, int*,
           int*, int*, int*);
extern "C" void dperm_(int*, REAL *, int *, int*, REAL *, 
                       int *, int *, int *, int *, int*);
extern "C" void vperm_(int*, REAL*, int*);
void genrcm ( int, int, int*, int*, int*);
void genmmd(int, int *, int *, int *, int *,
	    int, int *, int *, int *, int *,
	    int, int *);
/*----------------------------------------------------------*/
void symmgraph(struct csr_t *A, struct csr_t *C)
{
  int n = A->n;
  int nnz = A->nnz;

  struct csr_t AT;
  AT.ja = (int *) malloc(nnz*sizeof(int));
  AT.ia = (int *) malloc((n+1)*sizeof(int));
  // A': by csr2csc, job == 0 only pattern needed 
  csrcsc(n, n, 0, 1, NULL, A->ja, A->ia, NULL, AT.ja, AT.ia);
  
  // pattern of C = A+A'
  int nzmax = 2*nnz;
  C->ja = (int *) malloc(nzmax*sizeof(int));
  C->ia = (int *) malloc((n+1)*sizeof(int));
  int job = 0; // only the pattern is computed
  int *iw = (int *) malloc(n*sizeof(int));
  int err = -1;
  aplb_(&n, &n, &job, NULL, A->ja, A->ia, NULL, AT.ja,
        AT.ia, NULL, C->ja, C->ia, &nzmax, iw, &err);

  if (err != 0) {
    perror("SPARSKIT aplb error\n");
    exit(-1);
  }

  C->n = n;
  C->nnz = C->ia[n]-1;

  free(iw);
  free(AT.ja);
  free(AT.ia);
}

/*-------------------------------------------------*/
void Reorder_MMD(struct csr_t *A, int **perm, int **iperm)
{
  /* Minimum Degree */
  int mxint  = (1<<(8*sizeof(int)-2));
  struct csr_t C;
  symmgraph(A, &C);
  int n = C.n;
  int nnz = C.nnz;

  *perm  = (int *) malloc(n*sizeof(int));
  *iperm = (int *) malloc(n*sizeof(int));
  int delta = 1;
  int *head   = (int *) malloc(n*sizeof(int));
  int *qsize  = (int *) malloc(n*sizeof(int));
  int *Nlist  = (int *) malloc(n*sizeof(int));
  int *marker = (int *) malloc(n*sizeof(int));
  int nzout;
  genmmd(n, C.ia, C.ja, *perm,  *iperm, delta, head, qsize, 
         Nlist, marker,  mxint, &nzout);

  free(C.ia);
  free(C.ja);
  free(head);
  free(qsize);
  free(Nlist);
  free(marker);      
}

/*--------------------------------------------*/
void RowColPermMat(struct csr_t *A, int *perm)
{
  int job = 1;
  int n = A->n;
  int nnz = A->nnz;

  REAL *ao = (REAL *) malloc(nnz*sizeof(REAL));
  int *jao = (int *) malloc(nnz*sizeof(int));
  int *iao = (int *) malloc((n+1)*sizeof(int));

  dperm_(&n, A->a, A->ja, A->ia, 
             ao, jao, iao, perm, NULL, &job);

  free(A->a);
  free(A->ja);
  free(A->ia);
  A->a = ao;
  A->ia = iao;
  A->ja = jao;
}

/*---------------------------------------*/
void VecPerm(int n, REAL *x, int *perm)
{
  vperm_(&n, x, perm);
}
